/*******************************************************************************
 * 
 * jKiwi -- A virtual makeover and hairstyler application
 * Copyright (c) 2007-2008 Dan Mihai Ile, Maria Joao Roque Barbado Leal
 * 
 * This file is part of jKiwi.
 * 
 * jKiwi is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 * 
 * jKiwi is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 *******************************************************************************/

package dialogs;

import java.awt.Graphics2D;
import java.awt.RenderingHints;
import java.awt.image.BufferedImage;
import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.IOException;
import java.util.ResourceBundle;

import javax.imageio.ImageIO;
import javax.imageio.ImageReadParam;
import javax.imageio.ImageReader;
import javax.imageio.metadata.IIOMetadata;
import javax.imageio.metadata.IIOMetadataFormat;
import javax.imageio.stream.ImageInputStream;

import org.eclipse.swt.SWT;
import org.eclipse.swt.SWTException;
import org.eclipse.swt.custom.ScrolledComposite;
import org.eclipse.swt.events.DisposeEvent;
import org.eclipse.swt.events.DisposeListener;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.graphics.Rectangle;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Shell;
import org.w3c.dom.Node;

import com.drew.imaging.jpeg.JpegMetadataReader;
import com.drew.imaging.jpeg.JpegProcessingException;
import com.drew.metadata.Directory;
import com.drew.metadata.Metadata;
import com.drew.metadata.MetadataException;
import com.drew.metadata.exif.ExifDirectory;

import utils.ImageUtils;
import utils.StockIcon;

import core.Loader;

/**
 * Instances of this class show a browsing window
 * for images found at a given path
 */
public class PhotoBrowser {

	/**
	 * The shell used by this dialog
	 */
	private Shell sShell = null;

	/**
	 * Composite used to show the buttons to the bottom of the window
	 */
	private Composite compositeBottom = null;

	/**
	 * The close button
	 */
	private Button buttonClose = null;

	private final String IMAGES_PATH;

	private String selectedImage = null;

	private Label labelBefore = null;
	
	/**
	 * This dialog's Shell parent
	 */
	private final Shell parent;

	/**
	 * messages that are to be used by the entire GUI
	 */
	private ResourceBundle messages;

	//TODO check for the exceptions in the comment
	/**
	 * Creates a new instance of this class given it's parent Shell window
	 * @param parent the dialog's parent
	 *
	 * @exception IllegalArgumentException <ul>
	 *    <li>ERROR_NULL_ARGUMENT - if the parent is null</li>
	 * </ul>
	 * @exception SWTException <ul>
	 *    <li>ERROR_THREAD_INVALID_ACCESS - if not called from the thread that created the parent</li>
	 *    <li>ERROR_INVALID_SUBCLASS - if this class is not an allowed subclass</li>
	 * </ul>
	 */
	public PhotoBrowser(Shell parent, String imagesPath) {
		this.parent = parent;
		IMAGES_PATH = imagesPath;
		messages = Loader.getMessages();
		createSShell();
	}

	/**
	 * This method initializes sShell
	 */
	private void createSShell() {

		sShell = new Shell(parent, SWT.APPLICATION_MODAL | SWT.DIALOG_TRIM);
		sShell.setText(messages.getString("PHOTO_BROWSER_WINDOW_TITLE"));
		sShell.setLayout(new GridLayout());

		labelBefore = new Label(sShell, SWT.NONE);
		labelBefore.setText(messages.getString("PHOTO_BROWSER_WINDOW_LABEL"));

		createCompositeButtons();
		createCompositeBottom();

		//dispose the two created images on disposing this dialog
		sShell.addDisposeListener(new DisposeListener() {
			public void widgetDisposed(DisposeEvent e) {
			}
		});
	}

	public String open() {
		sShell.pack();
		sShell.open();
		sShell.setDefaultButton(buttonClose);
		buttonClose.setFocus();
		while (!sShell.isDisposed()) {
			if (!sShell.getDisplay().readAndDispatch())
				sShell.getDisplay().sleep();
		}
		return selectedImage;
	}

	/**
	 * This method initializes compositeAfter
	 */
	private void createCompositeButtons() {

		//grid data for the scrolled comoposite
		GridData gridDataScrolledComposite = new GridData();
		gridDataScrolledComposite.grabExcessVerticalSpace = true;
		gridDataScrolledComposite.heightHint = 300;
		gridDataScrolledComposite.verticalAlignment = org.eclipse.swt.layout.GridData.FILL;
		gridDataScrolledComposite.horizontalAlignment = org.eclipse.swt.layout.GridData.FILL;		

		//grid data for the scrolled composite container
		GridData gridDataScrolledCompositeContainer = new GridData();
		gridDataScrolledCompositeContainer.horizontalAlignment = org.eclipse.swt.layout.GridData.CENTER;
		gridDataScrolledCompositeContainer.grabExcessHorizontalSpace = true;

		//initialize the scrolled composite
		ScrolledComposite scrolledComposite = new ScrolledComposite(sShell, SWT.BORDER | SWT.V_SCROLL);
		scrolledComposite.setAlwaysShowScrollBars(true);
		scrolledComposite.setLayoutData(gridDataScrolledComposite);
		scrolledComposite.setExpandVertical(true);
		scrolledComposite.setExpandHorizontal(true);

		Composite compositeAfter = new Composite(scrolledComposite, SWT.NONE);
		GridLayout gl = new GridLayout();
		gl.numColumns = 4;
		compositeAfter.setLayout(gl);
		compositeAfter.setLayoutData(gridDataScrolledCompositeContainer);

		scrolledComposite.setContent(compositeAfter);
		
		File imagesDir = new File(IMAGES_PATH);

		for (final File f : imagesDir.listFiles()) {
			final File largeImgFile = new File(imagesDir.getAbsolutePath() + File.separator + f.getName());
			if (f.isFile() && isImage(f)) {
				try {
					GridData gd = new GridData();
					gd.widthHint = 100;
					gd.heightHint = 100;
					final Button b1 = new Button(compositeAfter, SWT.NONE);
					b1.setLayoutData(gd);
					sShell.getDisplay().asyncExec(new Runnable() {
						public void run() {
							BufferedImage resizedImage = null;
							if (!b1.isDisposed()) {
								BufferedImage bf = CreateImageFromFile(f);
								double size = ImageUtils.getResizingRatio(bf, 80, 80);
								resizedImage = resizeImage(bf, size, size);
								bf = null;
							}
							if (!b1.isDisposed()) {
								b1.setImage(new Image(Display.getCurrent(), ImageUtils.convertToSWT(resizedImage)));
							}
							resizedImage = null;
						}
					});

					b1.addSelectionListener(new org.eclipse.swt.events.SelectionAdapter() {
						public void widgetSelected(org.eclipse.swt.events.SelectionEvent e) {
							selectedImage = largeImgFile.getAbsolutePath();
							sShell.dispose();
						}
					});
					b1.setToolTipText(f.getName());

					// update the scrolled composite vertical scroll
					Rectangle r = scrolledComposite.getClientArea();
					scrolledComposite.setMinSize(compositeAfter.computeSize(r.width, SWT.DEFAULT));

				} catch (Exception e) {
					// skip the file
				}
			}
		}
	}
	
	/**
	 * Create a {@link BufferedImage} from a given file
	 * using subsampling for faster reading
	 * @param file the file to read
	 * @return the subsampled {@link BufferedImage}
	 */
	public BufferedImage CreateImageFromFile(File file) {
		
		//if jpg then try to read thumbail directly from metadata if available
		if (file.getName().toUpperCase().endsWith("JPG") ||
				file.getName().toUpperCase().endsWith("JPEG")) {
			try {
				// locate the bytes of the thumbnail data and read them using ImageIO
				// (thumbnail data will usually be a JPG image itself, if not there is a fallback)
				Metadata metadata = JpegMetadataReader.readMetadata(file);
				if (metadata.containsDirectory(ExifDirectory.class)) {
					Directory dir = metadata.getDirectory(ExifDirectory.class);
					if (dir.containsTag(ExifDirectory.TAG_THUMBNAIL_DATA)) {
						byte[] bytes = dir.getByteArray(ExifDirectory.TAG_THUMBNAIL_DATA);
						BufferedImage image = ImageIO.read(new ByteArrayInputStream(bytes));
						return image;
					}
				}
			} catch (JpegProcessingException ex) {
				ex.printStackTrace();
			} catch (MetadataException ex) {
				ex.printStackTrace();
			} catch (IOException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}
		
		//if no thumbnail found in metadata, resize the large image
		ImageReader reader = null;
		ImageReadParam param = null;
        ImageInputStream iis = null;
        BufferedImage img = null;
		try {
	        reader = (ImageReader)(ImageIO.getImageReadersByFormatName("jpeg").next());
	        param = reader.getDefaultReadParam();
			iis = ImageIO.createImageInputStream(file);
	        reader.setInput(iis,true);
	        param.setSourceSubsampling(10,10, 0, 0);
	        img = reader.read(0,param);
		} catch (IOException e1) {
	        reader = (ImageReader)(ImageIO.getImageReadersByFormatName("png").next());
	        param = reader.getDefaultReadParam();
			try {
				iis = ImageIO.createImageInputStream(file);
		        reader.setInput(iis,true);
		        param.setSourceSubsampling(10,10, 0, 0);
		        img = reader.read(0,param);
			} catch (IOException e) {
				// not jpg, not png, nothing to do
			}
		}
        return img;
    }

	/**
	 * Method that resize an given image using the width and height factor
	 * @param image the image that is to be resized
	 * @param widthFactor the width factor
	 * @param heightFactor the height factor
	 * @return the resized image
	 */
	private BufferedImage resizeImage(BufferedImage image, double widthFactor, double heightFactor) {

		BufferedImage firstPassImg = new BufferedImage((int)(image.getWidth() * widthFactor), (int)(image.getHeight() * heightFactor), BufferedImage.TYPE_INT_RGB);
		Graphics2D g2 = firstPassImg.createGraphics();
		g2.setRenderingHint(RenderingHints.KEY_RENDERING, RenderingHints.VALUE_RENDER_SPEED);
		g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_OFF);
		g2.setRenderingHint(RenderingHints.KEY_INTERPOLATION, RenderingHints.VALUE_INTERPOLATION_NEAREST_NEIGHBOR);
		g2.drawImage(image, 0, 0, firstPassImg.getWidth(), firstPassImg.getHeight(), null);
		g2.dispose(); 
		
		return firstPassImg;
		
//		AffineTransform tx = new AffineTransform(); //use affine transform class to do the resizing
//		tx.scale(widthFactor, heightFactor);
//		AffineTransformOp op = new AffineTransformOp(tx, AffineTransformOp.TYPE_NEAREST_NEIGHBOR);
//		return op.filter(image, null);
	}

	private boolean isImage(File f) {
		boolean res = false;
		if (f.getAbsolutePath().endsWith("jpg") || f.getAbsolutePath().endsWith("jpeg") || f.getAbsolutePath().endsWith("png")) {
			res = true;
		} else if (f.getAbsolutePath().endsWith("JPG") || f.getAbsolutePath().endsWith("JPEG") || f.getAbsolutePath().endsWith("PNG")) {
			res = true;
		}
		return res;
	}

	/**
	 * This method initializes compositeBottom
	 */
	private void createCompositeBottom() {
		GridData gridDataButtonClose = new GridData();
		GridLayout gridLayoutComposite = new GridLayout();
		gridLayoutComposite.marginHeight = 1;
		gridLayoutComposite.marginWidth = 1;
		GridData gridDataComposite = new GridData();
		gridDataComposite.horizontalSpan = 2;
		gridDataComposite.horizontalAlignment = org.eclipse.swt.layout.GridData.END;
		compositeBottom = new Composite(sShell, SWT.NONE);
		compositeBottom.setLayoutData(gridDataComposite);
		compositeBottom.setLayout(gridLayoutComposite);
		buttonClose = new Button(compositeBottom, SWT.NONE);
		buttonClose.setImage(new Image(Display.getCurrent(), StockIcon.getIcon(StockIcon.ICON_CLOSE, StockIcon.SIZE_LARGE_TOOLBAR, true)));
		buttonClose.setLayoutData(gridDataButtonClose);
		buttonClose.setText(messages.getString("PHOTO_BROWSER_WINDOW_CLOSE_BUTTON"));
		buttonClose.addSelectionListener(new org.eclipse.swt.events.SelectionAdapter() {
			public void widgetSelected(org.eclipse.swt.events.SelectionEvent e) {
				sShell.close();
			}
		});
		gridDataButtonClose.widthHint = buttonClose.computeSize(-1, -1).x + 10;
	}
}