/*******************************************************************************
 * 
 * jKiwi -- A virtual makeover and hairstyler application
 * Copyright (c) 2007-2008 Dan Mihai Ile, Maria Joao Roque Barbado Leal
 * 
 * This file is part of jKiwi.
 * 
 * jKiwi is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 * 
 * jKiwi is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 *******************************************************************************/

package tabs;

import java.awt.AlphaComposite;
import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.RenderingHints;
import java.awt.Stroke;
import java.awt.event.MouseEvent;
import java.awt.image.BufferedImage;

/**
 * Instances of this class can be used to
 * make line drawings onto the drawing area
 *
 */
public class ScribbleUtility {

	/**
	 * The stroke used to draw the lines
	 */
	private Stroke stroke = null;

	/**
	 * AlphaComposite use to specify the opacity that will be used
	 * to draw the currentDrawnContent on lastDrawingContent
	 */
	private AlphaComposite ac = null;

	/**
	 * Variables used to store mouse x,y starting and ending positions to draw the lines
	 */
	private int startX,startY,endX,endY;

	/**
	 * BufferedImage used to store the current drawn content
	 * (all the lines drawn until next mouse down event)
	 */
	private BufferedImage currentDrawnContent = null;

	/**
	 * The currentDrawnContent graphics used to draw on the image
	 */
	private Graphics2D currentDrawnContent_g2d = null;

	/**
	 * BufferedImage used to store all but the current drawn content
	 */
	private BufferedImage lastDrawingContent = null;

	/**
	 * The scribble's current drawing color
	 */
	private Color currentColor = null;

	//the width and height of the project's drawing image area
	private final int RESIZED_IMAGE_WIDTH;
	private final int RESIZED_IMAGE_HEIGHT;

	/**
	 * Creates a new instance of this class given the project's
	 * width and height of the drawing area
	 * @param resizedImageWidth width of the project's drawing area
	 * @param resizedImageHeight height of the project's drawing area
	 */
	public ScribbleUtility(int resizedImageWidth, int resizedImageHeight) {

		RESIZED_IMAGE_WIDTH = resizedImageWidth;
		RESIZED_IMAGE_HEIGHT = resizedImageHeight;
		lastDrawingContent = new BufferedImage(resizedImageWidth, resizedImageHeight, BufferedImage.TYPE_INT_ARGB);

	}

	/**
	 * Clear the scribble's image
	 */
	public void clear() {
		lastDrawingContent = new BufferedImage(RESIZED_IMAGE_WIDTH, RESIZED_IMAGE_HEIGHT, BufferedImage.TYPE_INT_ARGB);
	}

	/**
	 * Draw this scribble image using the given graphics context
	 * @param g2d the graphics context to use for drawing
	 */
	public void drawCurrentLayer(Graphics2D g2d) {

		g2d.drawImage(lastDrawingContent, 0, 0, null);
		java.awt.Composite c = g2d.getComposite(); //backup original g2d composite
		g2d.setComposite(ac); //user ConcealerTab composite
		g2d.drawImage(currentDrawnContent, 0, 0, null);
		g2d.setComposite(c); //restore g2d original composite

	}

	/**
	 * Returns a copy of the scribble's current image
	 * @return a copy of the scribble's current image
	 */
	public Object getCurrentState() {
		BufferedImage historyStep = new BufferedImage(RESIZED_IMAGE_WIDTH, RESIZED_IMAGE_HEIGHT, BufferedImage.TYPE_INT_ARGB);
		Graphics2D g2d_history = historyStep.createGraphics();
		g2d_history.drawImage(lastDrawingContent, 0, 0, null);
		g2d_history.dispose();
		return historyStep;
	}

	/**
	 * Returns the scribble's current image
	 * @return the scribble's current image
	 */
	public BufferedImage getLayerImage() {
		return lastDrawingContent;
	}

	/**
	 * Notify the scribble that the mouse was pressed;
	 * results in creating a new empty layer ready to
	 * be drawn with mouseDragged method
	 * @param e
	 */
	public void mousePressed(MouseEvent e) {

		endX = startX = e.getX();
		endY = startY = e.getY();

		currentDrawnContent = new BufferedImage(RESIZED_IMAGE_WIDTH, RESIZED_IMAGE_HEIGHT, BufferedImage.TYPE_INT_ARGB);

		currentDrawnContent_g2d = currentDrawnContent.createGraphics();
		currentDrawnContent_g2d.setRenderingHint(RenderingHints.KEY_ANTIALIASING,RenderingHints.VALUE_ANTIALIAS_ON);
		currentDrawnContent_g2d.setStroke(stroke);
		currentDrawnContent_g2d.setColor(currentColor);

		currentDrawnContent_g2d.drawLine(startX,startY,endX,endY);

	}

	/**
	 * Notify the scribble that the mouse was released;
	 * results in closing the current layer and add it
	 * to the scribble's final image
	 * @param e
	 */
	public void mouseReleased(MouseEvent e) {

		//update lastDrawingContent
		Graphics2D g2d = lastDrawingContent.createGraphics();
		g2d.setComposite(ac);
		g2d.drawImage(currentDrawnContent, 0, 0, null);
		g2d.dispose();

		//mouse is released, there is no currentDrawContent available
		currentDrawnContent_g2d.dispose();
		currentDrawnContent_g2d = null;
		currentDrawnContent = null;

	}

	/**
	 * Notify the scribble that the mouse is dragged;
	 * results in drawing a line in the scibble's
	 * current layer
	 * @param evt
	 */
	public void mouseDragged(MouseEvent evt) {

		endX = evt.getX();
		endY = evt.getY();
		currentDrawnContent_g2d.drawLine(startX,startY,endX,endY);
		startX = endX;
		startY = endY;

	}

	/**
	 * Set the current scribble's image
	 * @param o a BuferedImage containing the new scribble's image
	 */
	public void setCurrentState(Object o) {

		if (o instanceof BufferedImage) {
			BufferedImage b = (BufferedImage) o;
			lastDrawingContent = b;
		}

	}

	/**
	 * Change the scribble's drawing stroke
	 * @param stroke the new stroke to be used by the scribble
	 */
	public void setStroke(Stroke stroke) {
		this.stroke = stroke;
	}

	/**
	 * Change the scribble's drawing opacity
	 * @param ac the new opacity to be used by the scribble
	 */
	public void setAlphaOpacity(AlphaComposite ac) {
		this.ac = ac;
	}

	/**
	 * Change the scribble's drawing color
	 * @param color the new color to be used by the scribble
	 */
	protected void setCurrentColor(Color color) {
		currentColor = color;
	}

	/**
	 * Returns the scribble's current drawing color
	 * @return the scribble's current drawing color
	 */
	protected Color getcurrentColor() {
		return currentColor;
	}

	public void dispose() {
		currentDrawnContent = null;
		lastDrawingContent = null;
	}
}