/*******************************************************************************
 * 
 * jKiwi -- A virtual makeover and hairstyler application
 * Copyright (c) 2007-2008 Dan Mihai Ile, Maria Joao Roque Barbado Leal
 * 
 * This file is part of jKiwi.
 * 
 * jKiwi is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 * 
 * jKiwi is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 *******************************************************************************/

package tabs.hair;

import java.awt.image.BufferedImage;
import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.IOException;

import javax.imageio.ImageIO;


import org.eclipse.swt.SWT;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.events.DisposeEvent;
import org.eclipse.swt.events.DisposeListener;
import org.eclipse.swt.events.PaintEvent;
import org.eclipse.swt.events.PaintListener;
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.graphics.GC;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.graphics.RGB;
import org.eclipse.swt.widgets.Canvas;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Widget;
import org.xml.sax.SAXException;
import org.xml.sax.XMLReader;

import utils.Base64;

/**
 * Instances of this class represents a clickable button
 * with that contains the hair thumbnail
 * <p>
 * Note that although this class is a subclass of <code>Canvas</code>,
 * it does not make sense to add children to it, or set a layout on it.
 * </p>
 */
public class HairButton extends Canvas {

	//constants defining the button's colors
	private static final RGB BACKGROUND_COLOR = new RGB(255,255,255);
	private static final RGB NORMAL_BORDER_COLOR = new RGB(255,255,255);
	private static final RGB HOVER_BORDER_COLOR = new RGB(200,200,200);
	private static final RGB CLICKED_BORDER_COLOR = new RGB(150,150,150);
	private static final RGB SELECTED_BORDER_COLOR = new RGB(0,0,255);

	//the button's colors
	private Color backgroundColor = null;
	private Color normalBorderColor = null;
	private Color hoverBorderColor = null;
	private Color clickedBorderColor = null;
	private Color selectedBorderColor = null;

	/**
	 * Holds the current corder color which can be any of hover,normal,selected
	 */
	private Color currentBorderColor = null;

	/**
	 * The button's background image
	 */
	private Image backgroundImage = null;

	/**
	 * Flag to indicate whether this button is selected or not
	 */
	private boolean isSelected = false;

	/**
	 * Flag to indicate whether the background color should be disposed on widget dispose event
	 */
	private boolean isBackgroundColorDisposable = true;

	/**
	 * The XML hair file associated with this button
	 */
	private File associatedXML = null;

	/**
	 * Creates a new instance of this class given it's parent composite
	 * and a style value describing its behavior and appearance.
	 * <p>
	 * The style value is either one of the style constants defined in
	 * class <code>SWT</code> which is applicable to instances of this
	 * class, or must be built by <em>bitwise OR</em>'ing together 
	 * (that is, using the <code>int</code> "|" operator) two or more
	 * of those <code>SWT</code> style constants. The class description
	 * lists the style constants that are applicable to the class.
	 * Style bits are also inherited from superclasses.
	 * </p>
	 * 
	 * @param parent the parent for the button
	 * @param style the button's style
	 * 
	 * @see SWT
	 * @see Widget#getStyle
	 */
	public HairButton(Composite parent, int style) {
		super(parent, style);
		initialize();
	}

	/**
	 * initialize the hair button
	 */
	private void initialize() {
		setLayout(new GridLayout());

		//create the button's colors
		backgroundColor = new Color(Display.getCurrent(), BACKGROUND_COLOR);
		normalBorderColor = new Color(Display.getCurrent(), NORMAL_BORDER_COLOR);
		hoverBorderColor = new Color(Display.getCurrent(), HOVER_BORDER_COLOR);
		clickedBorderColor = new Color(Display.getCurrent(), CLICKED_BORDER_COLOR);
		selectedBorderColor = new Color(Display.getCurrent(), SELECTED_BORDER_COLOR);

		currentBorderColor = normalBorderColor;

		//dispose unnecessary resources when widget disposed
		addDisposeListener(new DisposeListener() {
			public void widgetDisposed(DisposeEvent e) {
				HairButton.this.widgetDisposed(e);
			}
		});

		//all the paint for the button get's done here
		addPaintListener(new PaintListener() {
			public void paintControl(PaintEvent e) {
				GC gc = e.gc;
				int width = getSize().x;
				int height = getSize().y;

				gc.setBackground(backgroundColor);
				gc.fillRectangle(0,0,width,height);
				if (backgroundImage != null) {
					gc.drawImage(backgroundImage, 0, 0, backgroundImage.getBounds().width, backgroundImage.getBounds().height, 1, 1, width - 2, height - 2);
				}
				updateBorder(gc);
			}
		});

		//mouse listeners for changing button state
		this.addMouseTrackListener(new org.eclipse.swt.events.MouseTrackAdapter() {   
			public void mouseExit(org.eclipse.swt.events.MouseEvent e) {
				currentBorderColor = normalBorderColor;
				updateBorder(null);
			}
			public void mouseEnter(org.eclipse.swt.events.MouseEvent e) {
				currentBorderColor = hoverBorderColor;
				updateBorder(null);
			}
		});
		this.addMouseListener(new org.eclipse.swt.events.MouseAdapter() {   
			public void mouseUp(org.eclipse.swt.events.MouseEvent e) {
				isSelected = true;
				updateBorder(null);
			}
			public void mouseDown(org.eclipse.swt.events.MouseEvent e) {
				currentBorderColor = clickedBorderColor;
				updateBorder(null);
				//TODO needed focus when there is a real mouse that clicks on the button (set focus for now is called in mouseDown listener of the creator class)
//				setFocus();
			}
		});
	}

	/**
	 * Change the selection state of this button
	 * @param isSelected the new button state
	 */
	public void setSelected(boolean isSelected) {
		this.isSelected = isSelected;
		if (!isSelected) {
			currentBorderColor = normalBorderColor;
		}
		updateBorder(null);
	}

	/**
	 * Update the button's border color using the given
	 * graphical context. If <code>null</code> is passed
	 * a new graphical context will be created for the drawing operations
	 * @param givenGc the graphical context to be used or <code>null</code>
	 */
	private void updateBorder(GC givenGc) {

		GC gc = null;
		if (givenGc == null) {
			gc = new GC(getAccessible().getControl());
		} else {
			gc = givenGc;
		}

		int width = getSize().x;
		int height = getSize().y;

		if (isSelected) {
			currentBorderColor = selectedBorderColor;
		}

		gc.setForeground(currentBorderColor);
		gc.drawRectangle(0,0,width-1,height-1);

		//dispose the created gc
		if (givenGc == null) {
			gc.dispose();
		}

	}

	/**
	 * Dispose the used resources on widget disposed event
	 */
	private void widgetDisposed(DisposeEvent e) {

		if (isBackgroundColorDisposable) {
			backgroundColor.dispose();
		}

		normalBorderColor.dispose();
		hoverBorderColor.dispose();
		clickedBorderColor.dispose();
		selectedBorderColor.dispose();

		associatedXML = null; //TODO necessary?

		//who should dispose the image is the caller for setBackgroundImage(Image image);
//		if (backgroundImage != null ) {
//		backgroundImage.dispose();
//		}

	}

	@Override
	public void setBackground(Color bkColor) {
		//the first time setBackground is called dispose the Color created in constructor
		if (isBackgroundColorDisposable) {
			backgroundColor.dispose();
		}
		isBackgroundColorDisposable = false;
		backgroundColor = bkColor;
	}

	@Override
	public Color getBackground() {
		return backgroundColor;
	}

	@Override
	public void setBackgroundImage(Image image) {
		backgroundImage = image;
	}

	@Override
	public Image getBackgroundImage() {
		return backgroundImage;
	}

	public void setAssociatedXML(File associatedXML) {
		this.associatedXML = associatedXML;
	}

	/**
	 * Return the hair images given a handler and XML parser
	 * @param handler hair style XML handler
	 * @param parser the XML reader parser
	 * @return an array of images containing the hair image, highlights image, shadow image
	 */
	public BufferedImage[] getImages(HairStyleXMLHandler handler, XMLReader parser) {
		BufferedImage highlightsImage = null;
		BufferedImage contentImage = null;
		BufferedImage shadowImage = null;
		parser.setContentHandler(handler);
		handler.initReadingVars(); //reset variables before reading
		try {
			parser.parse(associatedXML.toURI().toString());
		} catch (IOException e1) {
			// TODO Auto-generated catch block
			e1.printStackTrace();
		} catch (SAXException e1) {
			// TODO Auto-generated catch block
			e1.printStackTrace();
		}
		try {
			contentImage = ImageIO.read(new ByteArrayInputStream (Base64.decode(handler.getStrContent())));
			highlightsImage = ImageIO.read(new ByteArrayInputStream (Base64.decode(handler.getStrHighlights())));
			shadowImage = ImageIO.read(new ByteArrayInputStream (Base64.decode(handler.getStrShadow())));
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		return new BufferedImage[] {contentImage, highlightsImage, shadowImage};
	}
}